# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import logging
import shlex

import pmb.chroot
import pmb.config
import pmb.helpers.apk
import pmb.helpers.pmaports
import pmb.parse.apkindex
import pmb.parse.arch
import pmb.parse.depends
import pmb.parse.version


def update_repository_list(args, suffix="native", check=False):
    """
    Update /etc/apk/repositories, if it is outdated (when the user changed the
    --mirror-alpine or --mirror-pmOS parameters).

    :param check: This function calls it self after updating the
                  /etc/apk/repositories file, to check if it was successful.
                  Only for this purpose, the "check" parameter should be set to
                  True.
    """
    # Skip if we already did this
    if suffix in pmb.helpers.other.cache["apk_repository_list_updated"]:
        return

    # Read old entries or create folder structure
    path = f"{args.work}/chroot_{suffix}/etc/apk/repositories"
    lines_old = []
    if os.path.exists(path):
        # Read all old lines
        lines_old = []
        with open(path) as handle:
            for line in handle:
                lines_old.append(line[:-1])
    else:
        pmb.helpers.run.root(args, ["mkdir", "-p", os.path.dirname(path)])

    # Up to date: Save cache, return
    lines_new = pmb.helpers.repo.urls(args)
    if lines_old == lines_new:
        pmb.helpers.other.cache["apk_repository_list_updated"].append(suffix)
        return

    # Check phase: raise error when still outdated
    if check:
        raise RuntimeError(f"Failed to update: {path}")

    # Update the file
    logging.debug(f"({suffix}) update /etc/apk/repositories")
    if os.path.exists(path):
        pmb.helpers.run.root(args, ["rm", path])
    for line in lines_new:
        pmb.helpers.run.root(args, ["sh", "-c", "echo "
                                    f"{shlex.quote(line)} >> {path}"])
    update_repository_list(args, suffix, True)


def check_min_version(args, suffix="native"):
    """
    Check the minimum apk version, before running it the first time in the
    current session (lifetime of one pmbootstrap call).
    """

    # Skip if we already did this
    if suffix in pmb.helpers.other.cache["apk_min_version_checked"]:
        return

    # Skip if apk is not installed yet
    if not os.path.exists(f"{args.work}/chroot_{suffix}/sbin/apk"):
        logging.debug(f"NOTE: Skipped apk version check for chroot '{suffix}'"
                      ", because it is not installed yet!")
        return

    # Compare
    version_installed = installed(args, suffix)["apk-tools"]["version"]
    pmb.helpers.apk.check_outdated(
        args, version_installed,
        "Delete your http cache and zap all chroots, then try again:"
        " 'pmbootstrap zap -hc'")

    # Mark this suffix as checked
    pmb.helpers.other.cache["apk_min_version_checked"].append(suffix)


def install_build(args, package, arch):
    """
    Build an outdated package unless pmbootstrap was invoked with
    "pmbootstrap install" and the option to build packages during pmb install
    is disabled.

    :param package: name of the package to build
    :param arch: architecture of the package to build
    """
    # User may have disabled building packages during "pmbootstrap install"
    if args.action == "install" and not args.build_pkgs_on_install:
        if not pmb.parse.apkindex.package(args, package, arch, False):
            raise RuntimeError(f"{package}: no binary package found for"
                               f" {arch}, and compiling packages during"
                               " 'pmbootstrap install' has been disabled."
                               " Consider changing this option in"
                               " 'pmbootstrap init'.")
        # Use the existing binary package
        return

    # Build the package if it's in pmaports and there is no binary package
    # with the same pkgver and pkgrel. This check is done in
    # pmb.build.is_necessary, which gets called in pmb.build.package.
    return pmb.build.package(args, package, arch)


def packages_split_to_add_del(packages):
    """
    Sort packages into "to_add" and "to_del" lists depending on their pkgname
    starting with an exclamation mark.

    :param packages: list of pkgnames
    :returns: (to_add, to_del) - tuple of lists of pkgnames, e.g.
              (["hello-world", ...], ["some-conflict-pkg", ...])
    """
    to_add = []
    to_del = []

    for package in packages:
        if package.startswith("!"):
            to_del.append(package.lstrip("!"))
        else:
            to_add.append(package)

    return (to_add, to_del)


def packages_get_locally_built_apks(args, packages, arch):
    """
    Iterate over packages and if existing, get paths to locally built packages.
    This is used to force apk to upgrade packages to newer local versions, even
    if the pkgver and pkgrel did not change.

    :param packages: list of pkgnames
    :param arch: architecture that the locally built packages should have
    :returns: list of apk file paths that are valid inside the chroots, e.g.
              ["/mnt/pmbootstrap-packages/x86_64/hello-world-1-r6.apk", ...]
    """
    channel = pmb.config.pmaports.read_config(args)["channel"]
    ret = []

    for package in packages:
        data_repo = pmb.parse.apkindex.package(args, package, arch, False)
        if not data_repo:
            continue

        apk_file = f"{package}-{data_repo['version']}.apk"
        if not os.path.exists(f"{args.work}/packages/{channel}/{arch}/{apk_file}"):
            continue

        ret.append(f"/mnt/pmbootstrap-packages/{arch}/{apk_file}")

    return ret


def install_run_apk(args, to_add, to_add_local, to_del, suffix):
    """
    Run apk to add packages, and ensure only the desired packages get
    explicitly marked as installed.

    :param to_add: list of pkgnames to install, without their dependencies
    :param to_add_local: return of packages_get_locally_built_apks()
    :param to_del: list of pkgnames to be deleted, this should be set to
                   conflicting dependencies in any of the packages to be
                   installed or their dependencies (e.g. ["osk-sdl"])
    :param suffix: the chroot suffix, e.g. "native" or "rootfs_qemu-amd64"
    """
    # Sanitize packages: don't allow '--allow-untrusted' and other options
    # to be passed to apk!
    for package in to_add + to_add_local + to_del:
        if package.startswith("-"):
            raise ValueError(f"Invalid package name: {package}")

    commands = [["add"] + to_add]

    # Use a virtual package to mark only the explicitly requested packages as
    # explicitly installed, not the ones in to_add_local
    if to_add_local:
        commands += [["add", "-u", "--virtual", ".pmbootstrap"] + to_add_local,
                     ["del", ".pmbootstrap"]]

    if to_del:
        commands += [["del"] + to_del]

    for (i, command) in enumerate(commands):
        if args.offline:
            command = ["--no-network"] + command
        if i == 0:
            pmb.helpers.apk.apk_with_progress(args, ["apk"] + command,
                                              chroot=True, suffix=suffix)
        else:
            # Virtual package related commands don't actually install or remove
            # packages, but only mark the right ones as explicitly installed.
            # They finish up almost instantly, so don't display a progress bar.
            pmb.chroot.root(args, ["apk", "--no-progress"] + command,
                            suffix=suffix)


def install(args, packages, suffix="native", build=True):
    """
    Install packages from pmbootstrap's local package index or the pmOS/Alpine
    binary package mirrors. Iterate over all dependencies recursively, and
    build missing packages as necessary.

    :param packages: list of pkgnames to be installed
    :param suffix: the chroot suffix, e.g. "native" or "rootfs_qemu-amd64"
    :param build: automatically build the package, when it does not exist yet
                  or needs to be updated, and it is inside pmaports. For the
                  special case that all packages are expected to be in Alpine's
                  repositories, set this to False for performance optimization.
    """
    arch = pmb.parse.arch.from_chroot_suffix(args, suffix)

    if not packages:
        logging.verbose("pmb.chroot.apk.install called with empty packages list,"
                        " ignoring")
        return

    # Initialize chroot
    check_min_version(args, suffix)
    pmb.chroot.init(args, suffix)

    packages_with_depends = pmb.parse.depends.recurse(args, packages, suffix)
    to_add, to_del = packages_split_to_add_del(packages_with_depends)

    if build:
        for package in to_add:
            install_build(args, package, arch)

    to_add_local = packages_get_locally_built_apks(args, to_add, arch)
    to_add_no_deps, _ = packages_split_to_add_del(packages)

    logging.info(f"({suffix}) install {' '.join(to_add_no_deps)}")
    install_run_apk(args, to_add_no_deps, to_add_local, to_del, suffix)


def installed(args, suffix="native"):
    """
    Read the list of installed packages (which has almost the same format, as
    an APKINDEX, but with more keys).

    :returns: a dictionary with the following structure:
              { "postmarketos-mkinitfs":
                {
                  "pkgname": "postmarketos-mkinitfs"
                  "version": "0.0.4-r10",
                  "depends": ["busybox-extras", "lddtree", ...],
                  "provides": ["mkinitfs=0.0.1"]
                }, ...
              }
    """
    path = f"{args.work}/chroot_{suffix}/lib/apk/db/installed"
    return pmb.parse.apkindex.parse(path, False)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
""" Test pmb/parse/kconfig.py """
import pytest
import sys
import os

import pmb_test  # noqa
import pmb.parse.kconfig

test_options_checked_count = None


@pytest.fixture
def args(tmpdir, request):
    import pmb.parse
    sys.argv = ["pmbootstrap.py", "kconfig", "check"]
    args = pmb.parse.arguments()
    args.log = args.work + "/log_testsuite.txt"
    pmb.helpers.logging.init(args)
    request.addfinalizer(pmb.helpers.logging.logfd.close)
    return args


def patch_config(monkeypatch):
    """
    Delete the real kconfig_options_* variables in pmb/config/__init__.py and
    replace them with a very basic config for the tests. The idea is that it
    should use all features of the kconfig check code, so we can test all the
    code paths.
    """
    for key in list(pmb.config.__dict__.keys()):
        if key.startswith("kconfig_options"):
            monkeypatch.delattr(pmb.config, key)

    monkeypatch.setattr(pmb.config, "kconfig_options", {
        ">=0.0.0": {  # all versions
            "all": {  # all arches
                "ANDROID_PARANOID_NETWORK": False,
                "BLK_DEV_INITRD": True,
                "DEFAULT_HOSTNAME": "(none)",
            },
        },
        ">=2.6.0": {
            "all": {
                "BINFMT_ELF": True,
            },
        },
        "<4.7.0": {
            "all": {
                "DEVPTS_MULTIPLE_INSTANCES": True,
            },
        },
        "<5.2.0": {
            "armhf armv7 x86": {
                "LBDAF": True
            },
        },
    }, False)

    monkeypatch.setattr(pmb.config, "kconfig_options_waydroid", {
        ">=0.0.0": {
            "all": {
                "SQUASHFS": True,
                "ANDROID_BINDERFS": False,
                "ANDROID_BINDER_DEVICES": ["binder", "hwbinder", "vndbinder"],
            }
        },
    }, False)

    monkeypatch.setattr(pmb.config, "kconfig_options_nftables", {
        ">=3.13.0 <5.17": {
            "all": {
                "NFT_COUNTER": True,
            },
        },
    }, False)


def test_get_all_component_names(monkeypatch):
    patch_config(monkeypatch)
    func = pmb.parse.kconfig.get_all_component_names
    assert func() == ["waydroid", "nftables"]


def test_is_set():
    config = ("CONFIG_WIREGUARD=m\n"
              "# CONFIG_EXT2_FS is not set\n"
              "CONFIG_EXT4_FS=y\n")
    func = pmb.parse.kconfig.is_set
    assert func(config, "WIREGUARD") is True
    assert func(config, "EXT4_FS") is True
    assert func(config, "NON_EXISTING") is False


def test_is_set_str():
    config = 'CONFIG_DEFAULT_HOSTNAME="(none)"\n'
    func = pmb.parse.kconfig.is_set_str
    option = "DEFAULT_HOSTNAME"
    assert func(config, option, "(none)") is True
    assert func(config, option, "hello") is False
    assert func(config, f"{option}_2", "(none)") is False


def test_is_in_array():
    config = 'CONFIG_ANDROID_BINDER_DEVICES="binder,hwbinder,vndbinder"\n'
    func = pmb.parse.kconfig.is_in_array
    option = "ANDROID_BINDER_DEVICES"
    assert func(config, option, "binder") is True
    assert func(config, option, "hwbinder") is True
    assert func(config, option, "vndbinder") is True
    assert func(config, option, "invalidbinder") is False
    assert func(config, f"{option}_2", "binder") is False


def test_check_option():
    func = pmb.parse.kconfig.check_option
    config = ('CONFIG_BOOL=m\n'
              'CONFIG_LIST="a,b,c"\n'
              'CONFIG_STR="test"\n')
    path = "/home/user/myconfig.aarch64"

    assert func("test", False, config, path, "BOOL", True) is True
    assert func("test", True, config, path, "BOOL", True) is True
    assert func("test", True, config, path, "NON_EXISTING", True) is False
    assert func("test", True, config, path, "STR", "test") is True
    assert func("test", True, config, path, "STR", "test2") is False
    assert func("test", True, config, path, "LIST", ["a"]) is True
    assert func("test", True, config, path, "LIST", ["d"]) is False

    with pytest.raises(RuntimeError) as e:
        func("test", True, config, path, "TEST", {"dict": "notsupported"})
    assert "is not supported" in str(e.value)

    with pytest.raises(RuntimeError) as e:
        func("test", True, config, path, "TEST", None)
    assert "is not supported" in str(e.value)


def test_check_config_options_set():
    func = pmb.parse.kconfig.check_config_options_set
    config = ('CONFIG_BOOL=m\n'
              'CONFIG_LIST="a,b,c"\n'
              'CONFIG_STR="test"\n')
    path = "/home/user/myconfig.aarch64"
    arch = "aarch64"
    pkgver = "6.0"
    component = "testcomponent"

    # Skip check because version is too low
    options = {
        ">=6.0.1": {
            "all": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is True

    # Skip check because version is too high
    options = {
        "<6.0": {
            "all": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is True

    # Skip with two version that don't match
    options = {
        "<6.2 >=6.0.1": {
            "all": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is True

    # Version matches, arch does not match
    options = {
        ">=6.0": {
            "armhf": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is True

    # Version matches, arch matches (aarch64)
    options = {
        ">=6.0": {
            "aarch64": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is False

    # Version matches, arch matches (all)
    options = {
        ">=6.0": {
            "all": {
                "BOOL": False
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is False

    # Version matches, arch matches (all), rule passes
    options = {
        ">=6.0": {
            "all": {
                "BOOL": True
            }
        }
    }
    assert func(config, path, arch, options, component, pkgver) is True


def test_check_config_options_set_details(monkeypatch):
    global test_options_checked_count

    func = pmb.parse.kconfig.check_config_options_set
    config = ('CONFIG_BOOL=m\n'
              'CONFIG_LIST="a,b,c"\n'
              'CONFIG_STR="test"\n')
    path = "/home/user/myconfig.aarch64"
    arch = "aarch64"
    pkgver = "6.0"
    component = "testcomponent"

    def check_option_fake(*args, **kwargs):
        global test_options_checked_count
        test_options_checked_count += 1
        return False

    monkeypatch.setattr(pmb.parse.kconfig, "check_option", check_option_fake)

    options = {
        ">=0.0.0": {
            "all": {
                "BOOL": False,
                "STR": False,
            }
        }
    }

    # No details: stop after first error
    details = False
    test_options_checked_count = 0
    assert func(config, path, arch, options, component, pkgver, details) is False
    assert test_options_checked_count == 1

    # Details: don't stop, do both checks
    details = True
    test_options_checked_count = 0
    assert func(config, path, arch, options, component, pkgver, details) is False
    assert test_options_checked_count == 2


def test_check_config(monkeypatch, tmpdir):
    # Write test kernel config
    tmpdir = str(tmpdir)
    path = f"{tmpdir}/myconfig.aarch64"
    with open(path, "w") as handle:
        handle.write('CONFIG_BOOL=m\n'
                     'CONFIG_LIST="a,b,c"\n'
                     'CONFIG_STR="test"\n')

    patch_config(monkeypatch)

    func = pmb.parse.kconfig.check_config
    arch = "aarch64"
    pkgver = "6.0"

    # Invalid component
    components_list = ["invalid-component-name"]
    with pytest.raises(AssertionError) as e:
        func(path, arch, pkgver, components_list)
    assert "invalid kconfig component name" in str(e.value)

    # Fail base check
    components_list = []
    assert func(path, arch, pkgver, components_list) is False

    # Fails base check, even with enforce=False
    details = False
    enforce = False
    assert func(path, arch, pkgver, components_list, details, enforce) is False

    # Pass base check
    with open(path, "w") as handle:
        handle.write('CONFIG_BLK_DEV_INITRD=y\n'
                     'CONFIG_DEFAULT_HOSTNAME="(none)"\n'
                     'CONFIG_BINFMT_ELF=y\n')
    components_list = []
    assert func(path, arch, pkgver, components_list) is True

    # Fail additional check
    components_list = ["waydroid"]
    assert func(path, arch, pkgver, components_list) is False

    # Fail additional check, but result is still True with enforce=False
    components_list = ["waydroid"]
    details = True
    enforce = False
    assert func(path, arch, pkgver, components_list, details, enforce) is True


def test_check(args, monkeypatch, tmpdir):
    func = pmb.parse.kconfig.check
    details = True
    components_list = []
    patch_config(monkeypatch)

    # Create fake pmaports kernel structure
    tmpdir = str(tmpdir)
    monkeypatch.setattr(args, "aports", tmpdir)
    path_aport = f"{tmpdir}/device/community/linux-nokia-n900"
    path_apkbuild = f"{path_aport}/APKBUILD"
    os.makedirs(path_aport)

    # APKBUILD
    with open(path_apkbuild, "w") as handle:
        handle.write('pkgname=linux-nokia-n900\n'
                     'pkgver=5.15\n'
                     'options="pmb:kconfigcheck-nftables"\n')

    # Non-existing #1
    must_exist = True
    pkgname = "linux-does-not-exist"
    with pytest.raises(RuntimeError) as e:
        func(args, pkgname, components_list, details, must_exist)
    assert "Could not find aport" in str(e.value)

    # Non-existing #2
    must_exist = False
    pkgname = "linux-does-not-exist"
    assert func(args, pkgname, components_list, details, must_exist) is None

    # Invalid kernel config name
    path_kconfig = f"{path_aport}/config-nokia-n900_armv7"
    with open(path_kconfig, "w") as handle:
        handle.write('CONFIG_BOOL=m\n')
    must_exist = True
    pkgname = "linux-nokia-n900"
    with pytest.raises(RuntimeError) as e:
        func(args, pkgname, components_list, details, must_exist)
    assert "is not a valid kernel config" in str(e.value)
    os.unlink(path_kconfig)

    # Pass checks of base and nftables
    path_kconfig = f"{path_aport}/config-nokia-n900.armv7"
    with open(path_kconfig, "w") as handle:
        handle.write('CONFIG_BLK_DEV_INITRD=y\n'
                     'CONFIG_DEFAULT_HOSTNAME="(none)"\n'
                     'CONFIG_BINFMT_ELF=y\n'
                     'CONFIG_NFT_COUNTER=y\n')
    must_exist = True
    pkgname = "nokia-n900"
    assert func(args, pkgname, components_list, details, must_exist) is True

    # Don't pass nftables check
    with open(path_kconfig, "w") as handle:
        handle.write('CONFIG_BLK_DEV_INITRD=y\n'
                     'CONFIG_DEFAULT_HOSTNAME="(none)"\n'
                     'CONFIG_BINFMT_ELF=y\n')
    assert func(args, pkgname, components_list, details, must_exist) is False

    # Don't pass waydroid check (extra component check passed via cmdline)
    with open(path_kconfig, "w") as handle:
        handle.write('CONFIG_BLK_DEV_INITRD=y\n'
                     'CONFIG_DEFAULT_HOSTNAME="(none)"\n'
                     'CONFIG_BINFMT_ELF=y\n'
                     'CONFIG_NFT_COUNTER=y\n')
    components_list = ["waydroid"]
    assert func(args, pkgname, components_list, details, must_exist) is False


def test_extract_arch(tmpdir):
    func = pmb.parse.kconfig.extract_arch
    path = f"{tmpdir}/config"

    with open(path, "w") as handle:
        handle.write('CONFIG_ARM=y\n')
    assert func(path) == "armv7"

    with open(path, "w") as handle:
        handle.write('CONFIG_ARM64=y\n')
    assert func(path) == "aarch64"

    with open(path, "w") as handle:
        handle.write('CONFIG_RISCV=y\n')
    assert func(path) == "riscv64"

    with open(path, "w") as handle:
        handle.write('CONFIG_X86_32=y\n')
    assert func(path) == "x86"

    with open(path, "w") as handle:
        handle.write('CONFIG_X86_64=y\n')
    assert func(path) == "x86_64"

    with open(path, "w") as handle:
        handle.write('hello')
    assert func(path) == "unknown"


def test_extract_version(tmpdir):
    func = pmb.parse.kconfig.extract_version
    path = f"{tmpdir}/config"

    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# Linux/arm64 3.10.93 Kernel Configuration\n")
    assert func(path) == "3.10.93"

    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# Linux/arm64 6.2.0 Kernel Configuration\n")
    assert func(path) == "6.2.0"

    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# Linux/riscv 6.1.0-rc3 Kernel Configuration\n")
    assert func(path) == "6.1.0_rc3"

    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# no version here\n")
    assert func(path) == "unknown"


def test_check_file(tmpdir, monkeypatch):
    patch_config(monkeypatch)
    func = pmb.parse.kconfig.check_file
    path = f"{tmpdir}/config"

    # Fail the basic check
    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# Linux/arm64 3.10.93 Kernel Configuration\n"
                     "CONFIG_ARM64=y\n")

    func(path) is False

    # Pass the basic check
    with open(path, "w") as handle:
        handle.write("#\n"
                     "# Automatically generated file; DO NOT EDIT.\n"
                     "# Linux/arm64 3.10.93 Kernel Configuration\n"
                     "CONFIG_ARM64=y\n"
                     "BLK_DEV_INITRD=y\n"
                     "DEFAULT_HOSTNAME=\"(none)\"\n"
                     "BINFMT_ELF=y\n"
                     "DEVPTS_MULTIPLE_INSTANCES=y\n"
                     "LBDAF=y\n")

    func(path) is True



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import collections
import glob
import logging
import os
import shlex
import pmb.chroot
import pmb.helpers.cli


def get_ci_scripts(topdir):
    """ Find 'pmbootstrap ci'-compatible scripts inside a git repository, and
        parse their metadata (description, options). The reference is at:
        https://postmarketos.org/pmb-ci
        :param topdir: top directory of the git repository, get it with:
                       pmb.helpers.git.get_topdir()
        :returns: a dict of CI scripts found in the git repository, e.g.
                  {"flake8": {"description": "lint all python scripts",
                              "options": []},
                   ...} """
    ret = {}
    for script in glob.glob(f"{topdir}/.ci/*.sh"):
        is_pmb_ci_script = False
        description = ""
        options = []

        with open(script) as handle:
            for line in handle:
                if line.startswith("# https://postmarketos.org/pmb-ci"):
                    is_pmb_ci_script = True
                elif line.startswith("# Description: "):
                    description = line.split(": ", 1)[1].rstrip()
                elif line.startswith("# Options: "):
                    options = line.split(": ", 1)[1].rstrip().split(" ")
                elif not line.startswith("#"):
                    # Stop parsing after the block of comments on top
                    break

        if not is_pmb_ci_script:
            continue

        if not description:
            logging.error(f"ERROR: {script}: missing '# Description: …' line")
            exit(1)

        for option in options:
            if option not in pmb.config.ci_valid_options:
                raise RuntimeError(f"{script}: unsupported option '{option}'."
                                   " Typo in script or pmbootstrap too old?")

        short_name = os.path.basename(script).split(".", -1)[0]
        ret[short_name] = {"description": description,
                           "options": options}
    return ret


def sort_scripts_by_speed(scripts):
    """ Order the scripts, so fast scripts run before slow scripts. Whether a
        script is fast or not is determined by the '# Options: slow' comment in
        the file.
        :param scripts: return of get_ci_scripts()
        :returns: same format as get_ci_scripts(), but as ordered dict with
                  fast scripts before slow scripts """
    ret = collections.OrderedDict()

    # Fast scripts first
    for script_name, script in scripts.items():
        if "slow" in script["options"]:
            continue
        ret[script_name] = script

    # Then slow scripts
    for script_name, script in scripts.items():
        if "slow" not in script["options"]:
            continue
        ret[script_name] = script
    return ret


def ask_which_scripts_to_run(scripts_available):
    """ Display an interactive prompt about which of the scripts the user
        wishes to run, or all of them.
        :param scripts_available: same format as get_ci_scripts()
        :returns: either full scripts_available (all selected), or a subset """
    count = len(scripts_available.items())
    choices = ["all"]

    logging.info(f"Available CI scripts ({count}):")
    for script_name, script in scripts_available.items():
        extra = ""
        if "slow" in script["options"]:
            extra += " (slow)"
        logging.info(f"* {script_name}: {script['description']}{extra}")
        choices += [script_name]

    selection = pmb.helpers.cli.ask("Which script?", None, "all",
                                    complete=choices)
    if selection == "all":
        return scripts_available

    ret = {}
    ret[selection] = scripts_available[selection]
    return ret


def copy_git_repo_to_chroot(args, topdir):
    """ Create a tarball of the git repo (including unstaged changes and new
        files) and extract it in chroot_native.
        :param topdir: top directory of the git repository, get it with:
                       pmb.helpers.git.get_topdir() """
    pmb.chroot.init(args)
    tarball_path = f"{args.work}/chroot_native/tmp/git.tar.gz"
    files = pmb.helpers.git.get_files(args, topdir)

    with open(f"{tarball_path}.files", "w") as handle:
        for file in files:
            handle.write(file)
            handle.write("\n")

    pmb.helpers.run.user(args, ["tar", "-cf", tarball_path, "-T",
                                f"{tarball_path}.files"], topdir)

    ci_dir = "/home/pmos/ci"
    pmb.chroot.user(args, ["rm", "-rf", ci_dir])
    pmb.chroot.user(args, ["mkdir", ci_dir])
    pmb.chroot.user(args, ["tar", "-xf", "/tmp/git.tar.gz"],
                    working_dir=ci_dir)


def run_scripts(args, topdir, scripts):
    """ Run one of the given scripts after another, either natively or in a
        chroot. Display a progress message and stop on error (without printing
        a python stack trace).
        :param topdir: top directory of the git repository, get it with:
                       pmb.helpers.git.get_topdir()
        :param scripts: return of get_ci_scripts() """
    steps = len(scripts)
    step = 0
    repo_copied = False

    for script_name, script in scripts.items():
        step += 1

        where = "pmbootstrap chroot"
        if "native" in script["options"]:
            where = "native"

        script_path = f".ci/{script_name}.sh"
        logging.info(f"*** ({step}/{steps}) RUNNING CI SCRIPT: {script_path}"
                     f" [{where}] ***")

        if "native" in script["options"]:
            rc = pmb.helpers.run.user(args, [script_path], topdir,
                                      output="tui")
            continue
        else:
            # Run inside pmbootstrap chroot
            if not repo_copied:
                copy_git_repo_to_chroot(args, topdir)
                repo_copied = True

            env = {"TESTUSER": "pmos"}
            rc = pmb.chroot.root(args, [script_path], check=False, env=env,
                                 working_dir="/home/pmos/ci",
                                 output="tui")
        if rc:
            logging.error(f"ERROR: CI script failed: {script_name}")
            exit(1)



# SPDX-License-Identifier: GPL-3.0-or-later

import argparse

import pytest

from pmb.parse.arguments import toggle_other_boolean_flags


@pytest.fixture
def example_cli_with_flags():
    parser = argparse.ArgumentParser(prog="sample cli")
    parser.add_argument("-f1", "--flag1", action="store_true")
    parser.add_argument("-f2", "--flag2", action="store_true")
    return parser


def test_toggle_other_boolean_flags(example_cli_with_flags):
    other_flags = ["flag1", "flag2"]
    example_cli_with_flags.add_argument(
        "-f12", "--flag12",
        action=toggle_other_boolean_flags(*other_flags))
    args = example_cli_with_flags.parse_args(['-f12'])

    expected_flags_true = other_flags + ["flag12"]
    for flag in expected_flags_true:
        assert getattr(args, flag)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.aportgen.core
import pmb.helpers.git
import pmb.helpers.run


def generate(args, pkgname):
    # Copy original aport
    arch = pkgname.split("-")[1]
    upstream = pmb.aportgen.core.get_upstream_aport(args, "binutils")
    pmb.helpers.run.user(args, ["cp", "-r", upstream, args.work + "/aportgen"])

    # Rewrite APKBUILD
    fields = {
        "pkgname": pkgname,
        "pkgdesc": f"Tools necessary to build programs for {arch} targets",
        "arch": pmb.config.arch_native,
    }

    below_header = """
        CTARGET_ARCH=""" + arch + """
        CTARGET="$(arch_to_hostspec $CTARGET_ARCH)"
    """

    pmb.aportgen.core.rewrite(args, pkgname, "main/binutils", fields,
                              "binutils", below_header=below_header)
